import re
import shlex
import os
from shutil import which
from subprocess import check_output, Popen, PIPE

from .utils import does_path_exists
from .exceptions import YouTubeDLNotFound, DownloadOutPutDirDoesNotExits, DownloadFailed

"""
Python module to download the video from the input URL.
Uses youtube_dl or yt-dlp to download the video.
If both yt-dlp and youtube_dl are installed, it selects
yt-dlp. yt-dlp is more maintained than youtube_dl.
"""


class Download(object):
    """
    Class that downloads the video prior to frames extraction.

    Tries to download the lowest quality video possible.
    Uses youtube-dl or yt-dl to download the videos.
    """

    def __init__(self, url, output_dir, youtube_dl_path=None, worst=True):
        """
        Check if output_dir exists and is a directory, must end with "/".
        Check youtube_dl_path and verify that it is working else raises Exception.

        :param url: The URL of the video. The video will be
                    downloaded from this url. Must be a string.

        :param output_dir: The directory where the downloaded video will be stored.
                           Must be a string and path must be absolute.

        :param youtube_dl_path: If youtube_dl is not in path environment variable,
                                pass the path to the youtube-dl or yt-dlp to this param.
                                Must be string and only absolute path allowed.

        :param worst: The quality of video downloaded by the downloader.
                      True for worst quality and False for the default settings
                      of the downloader. The downloaders are yt-dlp and youtube_dl.
                      Default worst is True
        """
        self.url = url
        self.output_dir = output_dir
        self.youtube_dl_path = youtube_dl_path
        self.worst = worst

        if not does_path_exists(self.output_dir):
            raise DownloadOutPutDirDoesNotExits(
                "No directory called '%s' found for storing the downloaded video. Can not download the video."
                % self.output_dir
            )

        self._check_youtube_dl()

        self.download_video()

    def _check_youtube_dl(self):
        """Checks the youtube-dl and yt-dlp installations."""
        if not self.youtube_dl_path:

            if not which("yt-dlp"):

                if not which("youtube-dl"):

                    raise YouTubeDLNotFound(
                        "youtube-dl and yt-dlp not found on path. Install one of the two and add them to the path. Or you can also pass the path to 'youtube_dl_path' param"
                    )

                else:
                    self.youtube_dl_path = str(which("youtube-dl"))

            else:
                self.youtube_dl_path = str(which("yt-dlp"))

        try:

            output = check_output([str(self.youtube_dl_path), "--version"]).decode()

            if not re.search(
                r"[0-9]{4}\.[0-9]{2}\.[0-9]{2}", output
            ):  # raise Exception is youtube is not installed.
                # check by youtube-dl --version

                raise FileNotFoundError(
                    "The output is not matching the expected '%s --version' output."
                    % self.youtube_dl_path
                )

        except FileNotFoundError:

            raise YouTubeDLNotFound(
                "Youtube-dl/yt-dlp is not found at '%s'." % self.youtube_dl_path
            )

    def download_video(self):
        """Download the video from URL"""
        youtube_dl_path = self.youtube_dl_path
        output_dir = self.output_dir
        if os.name == "posix":
            youtube_dl_path = shlex.quote(self.youtube_dl_path)
            output_dir = shlex.quote(self.output_dir)

        worst = " "
        if self.worst:
            worst = " -f worst "

        command = (
            youtube_dl_path
            + worst
            + " "
            + '"'
            + self.url
            + '"'
            + " -o "
            + '"'
            + output_dir
            + "video_file.%(ext)s"
            + '"'
        )

        process = Popen(command, shell=True, stdout=PIPE, stderr=PIPE)
        output, error = process.communicate()

        youtube_dl_output = output.decode()
        youtube_dl_error = error.decode()

        if len(os.listdir(self.output_dir)) == 0:
            raise DownloadFailed(
                "%s failed to download the video at '%s'.\n%s\n%s"
                % (self.youtube_dl_path, self.url, youtube_dl_output, youtube_dl_error)
            )



import os
import random
import re
from PIL import Image
import imagehash
import shutil
from pathlib import Path

from .collagemaker import MakeCollage
from .downloader import Download
from .framesextractor import FramesExtractor
from .exceptions import DidNotSupplyPathOrUrl, StoragePathDoesNotExist
from .utils import (
    does_path_exists,
    create_and_return_temporary_directory,
    get_list_of_all_files_in_dir,
)


class VideoHash(object):
    """
    The VideoHash class provides an interface for computing & comparing the hash
    values for files supported by the ffmpeg. Every video format, encoding and
    containers that are supported by the ffmpeg can be used as input for this
    class.
    """

    def __init__(self, path=None, url=None, storage_path=None, download_worst=True):
        """
        :param path: absolute path of the video file.

        :param url: URL of the video file. Every URL that is supported by the
                    youtube-dl or yt-dlp package can be used.

        :param storage_path: If you want to provide a storage path for the files
                             created/downloaded by the instance, pass the
                             absolute path to that directory.

        :param download_worst: If set to False, download the default quality of
                               youtube-dl/yt-dlp downloader. These two downloaders
                               usually default to the best quality available.
                               Worst quality may be an issue for some users, they
                               are free to set the download_worst to False.
        """
        self.path = path
        self.url = url
        self.storage_path = storage_path
        self._storage_path = storage_path
        self.download_worst = download_worst
        self.video_path = None
        self.task_uid = VideoHash._get_task_uid()
        self._create_required_dirs_and_check_for_errors()
        self._copy_video_to_video_dir()
        FramesExtractor(self.video_path, self.frames_dir, interval=1, ffmpeg_path=None)
        self.collage_path = os.path.join(self.collage_dir, "collage.jpg")
        MakeCollage(
            get_list_of_all_files_in_dir(self.frames_dir),
            self.collage_path,
            collage_image_width=1024,
        )
        self.image = Image.open(self.collage_path)
        self.hash = None
        self.hash_hex = None
        self.bits_in_hash = 64

        self._calc_hash()

    def __str__(self):
        """
        The hash value of the instance. The hash value is 64 bit value prefixed
        with '0b', indicating the that the hash is binary.
        """

        return self.hash

    def __repr__(self):
        """
        Developer's representation of the instance.
        """

        return "VideoHash(hash=%s, hash_hex=%s, collage_path=%s, bits_in_hash=%s)" % (
            self.hash,
            self.hash_hex,
            self.collage_path,
            self.bits_in_hash,
        )

    def __len__(self):
        """
        Length of the hash value string. Total length is 66 characters, 64 for
        the bitstring and 2 for the prefix '0b' of the hash value string.
        """
        return len(self.hash)

    def __ne__(self, other):
        """
        Definition of the "!=" operator for the VideoHash objects.

        If the hamming distance of this instance and the other instance
        is zero returns False else returns True.
        """
        if self.__eq__(other):
            return False
        return True

    def __eq__(self, other):
        """
        Definition of the '=' operator on VideoHash objects.

        If the hamming distance of the instance and the other instance
        is zero returns True else returns False.
        """

        if self.__sub__(other) == 0:
            return True
        return False

    def __sub__(self, other):
        """
        Define how the '-' operator should work for the class.

        Checks that the binary strings are prefixed with '0b', hexadecimal
        strings prefixed with '0x' and if the string is not prefixed with
        any of them then raise ValueError.

        Instance of the VideoHash class are also supported.

        Raises ValueError if the object passed is not an instance of string
        nor VideoHash.
        """
        if other is None:
            raise TypeError("Other hash is None. And it should not be None.")

        if isinstance(other, str):
            if other.lower().startswith("0x"):
                return VideoHash.hamming_distance(
                    self.hash, VideoHash.hex2bin(other.lower(), self.bits_in_hash)
                )
            elif other.lower().startswith("0b"):
                if len(other) != len(self.hash):
                    raise ValueError(
                        "Can not compare different bits hashes. You must supply a %d bits hash."
                        % self.bits_in_hash
                    )
                return VideoHash.hamming_distance(self.hash, other.lower())
            else:
                raise TypeError(
                    "Hash string must start with either '0x' for hexadecimal or '0b' for binary."
                )

        if isinstance(other, VideoHash):
            return VideoHash.hamming_distance(self.hash, other.hash)

        raise TypeError(
            "To calculate difference both of the hashes must be either hexadecimal/binary strings or instance of VideoHash"
        )

    def _copy_video_to_video_dir(self):
        """
        Copy the video from the path the video directory.
        It avoids the issues like the user or some other
        processes deleting the instance files while we are
        still processing.

        If instead of the path the uploader specified an url,
        then download the video and copy the file to video
        directory.

        :raises ValueError: If the path supplied by the end user
                            lacks an extension. Like webm or mp4.
        """
        if self.path:
            # create a copy of the video at self.storage_path
            match = re.search(r"\.([^.]+$)", self.path)

            if match:
                extension = match.group(1)
            else:
                raise ValueError("File name (path) does not have an extension.")

            self.video_path = os.path.join(self.video_dir, ("video.%s" % extension))
            shutil.copyfile(self.path, self.video_path)

        if self.url:
            Download(
                self.url,
                self.video_download_dir,
                youtube_dl_path=None,
                worst=self.download_worst,
            )
            downloaded_file = get_list_of_all_files_in_dir(self.video_download_dir)[0]
            self.video_path = "%svideo.%s" % (
                self.video_dir,
                re.search(r"\.(.*?)$", downloaded_file).group(1),
            )
            shutil.copyfile(downloaded_file, self.video_path)

    def _create_required_dirs_and_check_for_errors(self):
        """
        Creates some directories before the main processing starts.
        The files are stored in these directories, no need to worry about
        the end user or some other processes interfering with the instance
        generated files.

        :raises DidNotSupplyPathOrUrl: If the user forgot to specify both the
                                       path and the url. One of them must be
                                       specified.

        :raises ValueError: If user passed both the path and the url. Only pass
                            one of them if the file is available on both the
                            path and the url.

        :raises StoragePathDoesNotExist: If the storage path specified by the
                                         user in does not exists.
        """
        if not self.path and not self.url:
            raise DidNotSupplyPathOrUrl(
                "You must specify either a path or an URL of the video."
            )

        if self.path and self.url:
            raise ValueError("Specify either a path or an URL and NOT both.")

        if not self.storage_path:
            self.storage_path = create_and_return_temporary_directory()

        if not does_path_exists(self.storage_path):
            raise StoragePathDoesNotExist(
                "Storage path '%s' does not exist." % self.storage_path
            )

        self.storage_path = os.path.join(
            self.storage_path, ("%s%s" % (self.task_uid, os.path.sep))
        )
        self.video_dir = os.path.join(self.storage_path, ("video%s" % os.path.sep))
        Path(self.video_dir).mkdir(parents=True, exist_ok=True)
        self.video_download_dir = os.path.join(
            self.storage_path, ("downloadedvideo%s" % os.path.sep)
        )
        Path(self.video_download_dir).mkdir(parents=True, exist_ok=True)
        self.frames_dir = os.path.join(self.storage_path, ("frames%s" % os.path.sep))
        Path(self.frames_dir).mkdir(parents=True, exist_ok=True)
        self.collage_dir = os.path.join(self.storage_path, ("collage%s" % os.path.sep))
        Path(self.collage_dir).mkdir(parents=True, exist_ok=True)

    def delete_storage_path(self):
        """Delete the storage_path directory tree."""
        dir = self.storage_path
        if not self._storage_path:
            dir = "%s%s" % (
                os.path.dirname(os.path.dirname(os.path.dirname(self.storage_path))),
                os.path.sep,
            )
        shutil.rmtree(dir, ignore_errors=True, onerror=None)

    @staticmethod
    def _get_task_uid():
        """
        Returns an unique task id for the instance and the task id is used to
        differentiate the instance files from the other unrelated files.
        """
        sys_random = random.SystemRandom()
        return "".join(
            sys_random.choice("abcdefghijklmnopqrstuvwxyz0123456789") for _ in range(12)
        )

    @staticmethod
    def hamming_distance(string_a, string_b):
        """
        Compute the hamming distance of the input strings.

        :raises ValueError: The input strings are of unequal length. Hamming
                            distance is not defined for unequal length strings.
        """
        if len(string_a) != len(string_b):
            raise ValueError(
                "Strings are of unequal length can not compute hamming distance. Hamming distance is undefined."
            )
        return sum(char_1 != char_2 for char_1, char_2 in zip(string_a, string_b))

    @staticmethod
    def hex2bin(hexstr, padding):
        """
        Convert hexadecimal('0x' prefixed) to bitstring string prefixed
        with '0b'.

        :raises ValueError: It input hexadecimal string is not prefixed
                            with '0x'.
        """
        if not hexstr.lower().startswith("0x"):
            raise ValueError("Input hexadecimal string must have '0x' as the prefix.")
        return "0b%s" % (
            str(bin(int(hexstr.lower(), 0))).replace("0b", "").zfill(padding)
        )

    @staticmethod
    def bin2hex(binstr):
        """
        Convert bitstring('0b' prefixed) to hexadecimal string prefixed
        with '0x'.

        :raises ValueError: It input binary string is not prefixed with '0b'.
        """

        if not binstr.lower().startswith("0b"):
            raise ValueError("Binary string must be prefixed with '0b'.")

        return str(hex(int(binstr, 2)))

    def _calc_hash(self):
        """
        Calculates the binary hash value by calling the whash method of
        imagehash package.

        End-user is not provided any access to the imagehash instance but
        instead the binary and hexadecimal equivalent of the result of
        wavelet-hash.
        """

        wavelethash_bit_list = []
        for row in imagehash.whash(self.image).hash.tolist():
            wavelethash_bit_list.extend(row)

        self.hash = ""

        for bit in wavelethash_bit_list:
            if bit:
                self.hash += "1"
            else:
                self.hash += "0"

        # the binary value is prefixed with 0b.
        self.hash = "0b%s" % self.hash
        self.hash_hex = VideoHash.bin2hex(self.hash)



__title__ = "videohash"
__description__ = "Python package for Perceptual Video Hashing (Near Duplicate Video Detection) - Get a 64-bit comparable hash-value for any video."
__url__ = "https://akamhy.github.io/videohash/"
__version__ = "2.0.2"
__status__ = "production"
__author__ = "Akash Mahanty"
__author_email__ = "akamhy@yahoo.com"
__license__ = "MIT"
__copyright__ = "Copyright 2021 Akash Mahanty"



"""
videohash.exceptions
~~~~~~~~~~~~~~~~~~~
This module contains the videohash's exceptions.
"""


class VideoHashError(Exception):
    """Main Exception for the package"""

    pass


class DownloadError(Exception):
    """Base for all downloading errors."""

    pass


class FFmpegError(Exception):
    """Base FFmpeg software error"""

    pass


class DownloadSoftwareError(DownloadError):
    """Base for all youtube_dl and yt-dlp errors."""

    pass


class DownloadFailed(DownloadError):
    """Failed to download the video file from url supplied"""

    pass


class DownloadOutPutDirDoesNotExits(DownloadError):
    """The output directory supplied to store the downloaded video does not exists."""

    pass


class YouTubeDLNotFound(DownloadSoftwareError):
    """youtube-dl and yt-dl are not found in path"""

    pass


class FFmpegNotFound(FFmpegError):
    """FFmpeg is not is the path"""

    pass


class FFmpegFailedToExtractFrames(FFmpegError):
    """FFmpeg failed to extract any frame"""

    pass


class FramesExtractorOutPutDirDoesNotExits(Exception):
    """The output dir passed to the frame extractor does not exits."""

    pass


class StoragePathDoesNotExist(Exception):
    """
    The supplied storage path does not exists.
    Do not supply if you don't care about the collage or extracted farmes.
    The collage is the image representing your video.
    """

    pass


class DidNotSupplyPathOrUrl(Exception):
    """Must supply either a path for your video or a valid URL"""

    pass


class CollageOfZeroFramesError(Exception):
    """Raised if zero frames are passed for collage making"""

    pass



import os
import shlex
from shutil import which
from subprocess import check_output, Popen, PIPE

from .utils import does_path_exists
from .exceptions import (
    FramesExtractorOutPutDirDoesNotExits,
    FFmpegNotFound,
    FFmpegFailedToExtractFrames,
)


"""
python module to extract the frames from the input video.
Uses the FFmpeg Software to extract the frames.
"""


class FramesExtractor(object):
    """
    extract from from the passed video file and save at the output directory.
    """

    def __init__(self, video_path, output_dir, interval=1, ffmpeg_path=None):
        """
        Raises Exeception if video_path does not exists.
        Raises Exeception if output_dir does not exists or if not a directory.
        Checks  the ffmpeg installation and the path; thus ensure that we can use it.

        :param video_path: absolute path of the video

        :param output_dir: absolute path of the directory
                           where to save the frames.

        :param interval: interval is seconds. interval must be an integer.
                         Extract one frame every given number of seconds.
                         Default is 1, that is one frame every second.

        :param ffmpeg_path: path of the ffmpeg software if not in path.

        """
        self.video_path = video_path
        self.output_dir = output_dir
        self.interval = interval
        self.ffmpeg_path = ffmpeg_path

        if not does_path_exists(self.video_path):
            raise FileNotFoundError(
                "No video found at '%s' for frame extraction." % self.video_path
            )

        if not does_path_exists(self.output_dir):
            raise FramesExtractorOutPutDirDoesNotExits(
                "No directory called '%s' found for storing the frames."
                % self.output_dir
            )

        self._check_ffmpeg()

        self.extract()

    def _check_ffmpeg(self):
        """
        Checks the ffmpeg path and runs 'ffmpeg -version' to verify that the
        software, ffmpeg is found and works.
        """
        if not self.ffmpeg_path:
            if not which("ffmpeg"):
                raise FFmpegNotFound(
                    "FFmpeg is not on the path. Install FFmpeg and add it to the path."
                    + "Or you can also pass the path via the 'ffmpeg_path' param."
                )
            else:
                self.ffmpeg_path = str(which("ffmpeg"))

        # Check the ffmpeg
        try:
            # check_output will raise FileNotFoundError if it does not finds the ffmpeg
            output = check_output([str(self.ffmpeg_path), "-version"]).decode()
        except FileNotFoundError:
            raise FFmpegNotFound("FFmpeg not found at '%s'." % self.ffmpeg_path)
        else:
            if "ffmpeg version" not in output:
                raise FFmpegNotFound(
                    "ffmpeg at '%s' is not really ffmpeg. Output of ffmpeg -version is \n'%s'."
                    % (self.ffmpeg_path, output)
                )

    def extract(self):
        """
        Extract the frames at every n seconds where n is the
        integer set to self.interval.
        """
        ffmpeg_path = self.ffmpeg_path
        video_path = self.video_path
        output_dir = self.output_dir
        if os.name == "posix":
            ffmpeg_path = shlex.quote(self.ffmpeg_path)
            video_path = shlex.quote(self.video_path)
            output_dir = shlex.quote(self.output_dir)

        command = (
            ffmpeg_path
            + " -i "
            + '"'
            + video_path
            + '"'
            + " -s 144x144 "
            + " -r "
            + str(self.interval)
            + " "
            + '"'
            + output_dir
            + "video_frame_%07d.jpeg"
            + '"'
        )

        process = Popen(command, shell=True, stdout=PIPE, stderr=PIPE)
        output, error = process.communicate()

        ffmpeg_output = output.decode()
        ffmpeg_error = error.decode()

        if len(os.listdir(self.output_dir)) == 0:
            raise FFmpegFailedToExtractFrames(
                "FFmpeg could not extract any frames.\n%s\n%s\n%s"
                % (command, ffmpeg_output, ffmpeg_error)
            )



from math import ceil, sqrt
from PIL import Image
import os

from .exceptions import CollageOfZeroFramesError
from .utils import does_path_exists

"""
Module to create collage from list of images, the
images are the extracted frames of the input video.
"""


class MakeCollage(object):
    """
    Class that creates the collage from list of images.

    Collage that should be as close to the shape of a square.

    The images are arranged by timestamp of the frames, their
    index in the image_list is based on thier timestamp on the
    video.

    Example:
    Let's say we have a list with 9 images.
    As the images should be arranged in a way to resemble a
    square, we take the square root of 9 and that is 3. Now
    we need to make a 3x3 frames collage.

    Arrangement should be:
    Img1 Img2 Img3
    Img4 Img5 Img6
    Img7 Img8 Img9

    If the number of images is not a perfect square, calculate the
    square root and round it to the nearest integer.

    If number of images is 13, which is not a perfect square.

    sqrt(13) = 3.605551275463989
    round(3.605551275463989) = 4

    Thus the image should be 4x4 frames of collage.

    Arrangement should be:
    -----------------------------
    |  Img1  Img2  Img3   Img4  |
    |  Img5  Img6  Img7   Img8  |
    |  Img9  Img10 Img11  Img12 |
    |  Img13  X     X      X    |
    -----------------------------

    X denotes the empty space due to lack of images.
    But the empty spaces will not affect the robustness
    as downsized version of the video will also have these
    vacant spaces.
    """

    def __init__(self, image_list, output_path, collage_image_width=1024):
        """
        Checks if the list passed is not an empty list.
        Also makes sure that the output_path directory exists.

        And calls the make method, make method creates the collage.

        :param image_list: A python list instance containing the list of absolute
                           path of images that are to be added in the collage.
                           The order of images is kept intact.

        :param output_path: A string of the absolute path of the image including
                            the image name.
                            Example: '/home/username/projects/collage.jpeg'.

        :param collage_image_width: An integer specifying the image width of the
                                    output collage. Default value is 1024 pixels.
        """
        self.image_list = image_list
        self.number_of_images = len(self.image_list)
        self.output_path = output_path
        self.collage_image_width = collage_image_width

        # The algorithm will calculate the collage image height and set
        # the value to this attribute.
        self.collage_image_height = None

        self.images_per_row_in_collage = int(round(sqrt(self.number_of_images)))

        if self.number_of_images == 0:
            raise CollageOfZeroFramesError("Can not make a collage of zero images.")

        output_path_dir = os.path.dirname(self.output_path) + "/"
        if not does_path_exists(output_path_dir):
            raise FileNotFoundError(
                "Directory at which output collage is to be saved does not exists."
            )

        self.make()

    def make(self):
        """
        Creates the collage from list of images.

        It calculates the scale of the images on collage by
        measuring the first image width and height, there's no
        reason for choosing first one and it's arbitrary. But
        we assume that all the images passed should have same size.

        Read the comments made in the code to understand how the
        collage maker algorithm works.
        """
        # arbitrarily selecting the first image from the list, index 0
        first_frame_image = Image.open(self.image_list[0])

        # calculate the width and height of the first image of the list.
        # Here we assume that all the images passed should have same size.
        frame_image_width, frame_image_height = first_frame_image.size

        # scale is the ratio of collage_image_width and product of
        # images_per_row_in_collage with frame_image_width.
        # clearly denominator will be bigger than numerator unless
        # collage_image_width is set to a very small integer, the video
        # is of very low resolution or collage_image_width is set to a big
        # integer.

        # Therefore scale will always lie between 0 and 1, which implies that
        # the images are always going to get downsized.
        scale = (self.collage_image_width) / (
            self.images_per_row_in_collage * frame_image_width
        )

        # scaled frame image width and height are the height and width
        # they will occupy on the collage.
        scaled_frame_image_width = ceil(frame_image_width * scale)
        scaled_frame_image_height = ceil(frame_image_height * scale)

        # Divide the number of images by images_per_row_in_collage. The later
        # was calculated by taking the square root of total images.
        number_of_rows = ceil(self.number_of_images / self.images_per_row_in_collage)

        # we are multiplying the height of one downsized image with number of rows.
        # height of 1 downsized image = scale * frame_image_height
        # total height is clearly the multiplication of number of rows and height of
        # one downsized image.
        self.collage_image_height = ceil(scale * frame_image_height * number_of_rows)

        # Create an image of passed collage_image_width and calculated collage_image_height.
        # The downsized images will be pasted on this new base image.
        # The image is 0,0,0 RGB(black) and can't affect the hash value also it reduces the
        # issues with the black-bars in some videos.
        collage_image = Image.new(
            "RGB", (self.collage_image_width, self.collage_image_height)
        )

        # keep track of the x and y coordinates
        i, j = (0, 0)

        # iterate the frames and paste them on their position on the collage_image
        for count, frame in enumerate(self.image_list):

            # Set the x coordinate to zero if we are on the first column
            # If self.images_per_row_in_collage is 4
            # then 0,4,8 and so on should have their x coordinate as 0
            if (count % self.images_per_row_in_collage) == 0:
                i = 0

            # open the image, must open it to resize it using the thumbnail method
            frame = Image.open(frame)

            # scale the opened images
            frame.thumbnail(
                (scaled_frame_image_width, scaled_frame_image_height), Image.ANTIALIAS
            )

            # set the value x to that of i's value.
            x = i

            # Set the value of y to the row number of a zero based row indexing
            # If images_per_row_in_collage is 4 then upto 4 but not including it
            # we will get a zero. zero = first row's y coordinate.
            # from 4 to 7 it will set the value of y to 2 and so on.
            # It ensure that y coordinate stays the same for any given row.
            y = (j // self.images_per_row_in_collage) * scaled_frame_image_height

            # paste the image on the newly created base image
            collage_image.paste(frame, (x, y))

            # increase the x coordinate by scaled_frame_image_width
            # to get the x coordinate of the next frame. unless the next image
            # will be on the very first column this will be the x coordinate
            i = i + scaled_frame_image_width

            # increase the value of j by 1, this is to get calculate y coordinate of
            # next image. The increase number will be floor divided by images_per_row_in_collage
            # there for the y stays the same for any given row.
            j += 1

        # save the base image with all the scaled images pasted on it.
        collage_image.save(self.output_path)
